# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Rust's `Option<T> error handling implementation for Python`."""

from __future__ import annotations

__all__: tuple[str, ...] = ("Some", "Fn", "FnOnce", "ValueT", "OkT", "ErrT", "Option")

import typing

from . import default as _default
from . import ref as _ref

ValueT = typing.TypeVar("ValueT")
"""A type hint that represents the generic value of the `Some` type."""
OkT = typing.TypeVar("OkT")
"""The `Ok` result of a `Result`."""
ErrT = typing.TypeVar("ErrT")
"""The `Err` result of a `Result`."""

T = typing.TypeVar("T")
T_co = typing.TypeVar("T_co", covariant=True)

Fn = typing.Callable[[ValueT], T]
"""A type hint for a function that can take a `ValueT` and return a `T`."""
FnOnce = typing.Callable[[], T]
"""A type hint for a function that takes no arguments and return `T`"""


@typing.final
class Some(typing.Generic[ValueT], _default.Default[None]):
    """The `Some` object. A typesafe object is used for values that might be `T` or `None`.

    It is similar to `typing.Optional[T]`, But has proper methods to handle the contined value.

    Example
    -------
    ```py
    value = Some("Hello")
    print(value)
    # Some("Hello")

    # This will unwrap the contained value as long as
    # it is not `None` otherwise this will raise an error.
    print(value.unwrap())
    # "Hello"

    none_value = Some(None)
    while none_value.unwrap():
        # Never unreachable!

    # Solving it with `unwrap_or` method to unwrap the value or return a default value.
    print(none_value.unwrap_or(10))
    # 10
    ```
    """

    __slots__ = ("_value",)

    def __init__(self, value: ValueT | None, /) -> None:
        self._value = value

    @staticmethod
    def default() -> None:
        """Return the default value for `Some`. This always returns `None`."""
        return None

    @property
    def read(self) -> ValueT | None:
        """Read the contained value."""
        return self._value

    def unwrap(self) -> ValueT:
        """Unwrap the inner value either returning if its not `None` or raising a `RuntimeError`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.unwrap())
        # 5

        value = Some(None)
        print(value.unwrap())
        # RuntimeError
        ```

        Raises
        ------
        `RuntimeError`
            If the inner value is `None`.
        """
        if self._value is None:
            raise RuntimeError(f"Can't unwrap {type(self).__name__} value.") from None

        return self._value

    def unwrap_or(self, default: ValueT, /) -> ValueT:
        """Unwrap the inner value either returning if its not `None` or returning `default`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.unwrap_or(10))
        # 5

        # Type hint is required here.

        value: Some[int] = Some(None)
        print(value.unwrap_or(10))
        # 10
        ```
        """
        if self._value is None:
            return default

        return self._value

    def unwrap_or_else(self, f: FnOnce[ValueT], /) -> ValueT:
        """Unwrap the inner value either returning if its not `None` or calling `f` to get a default value.

        Example
        -------
        ```py
        value = Some(5)
        print(value.unwrap_or_else(lambda: 10))
        # 5

        value: Some[bool] = Some(None)
        print(value.unwrap_or_else(lambda: True))
        # True
        ```
        """
        if self._value is None:
            return f()

        return self._value

    # Functional.
    def map(self, f: Fn[ValueT, T], /) -> Some[T | None]:
        """Map the inner value to a new value. Returning `Some[None]` if `ValueT` is `None`.

        Example
        -------
        ```py
        value = Some(5.0)

        print(value.map(lambda x: x * 2.0))
        # Some(10.0)

        value: Some[bool] = Some(None)
        print(value)
        # Some(None)
        ```
        """
        if self._value is None:
            return Some(None)

        return Some(f(self._value))

    def map_or(self, default: T, f: Fn[ValueT, T], /) -> T:
        """Map the inner value to a new value or return `default` if its `None`.

        Example
        -------
        ```py
        value = Some(5.0)

        # Since the value is not `None` this will get mapped.
        print(value.map_or(10.0, lambda x: x + 1.0))
        # 6.0

        # This is `None`, so the default value will be returned.
        value: Some[str] = Some(None)
        print(value.map_or("5", lambda x: str(x)))
        # "5"
        ```
        """
        if self._value is None:
            return default

        return f(self._value)

    def map_or_else(self, default: FnOnce[T], f: typing.Callable[[ValueT], T], /) -> T:
        """Map the inner value to a new value, Or return default which maps to `default()` if its `None`.

        Example
        -------
        ```py
        value = Some(5.0)

        # Since the value is not `None` this will get mapped.
        print(value.map_or_else(lambda: 10.0, lambda x: x + 1.0))
        # 6.0

        # This is `None`, so the default func will be returned.
        value: Option[str] = Some(None)
        print(value.map_or_else(lambda: "5", lambda x: str(x)))
        # "5"
        ```
        """
        if self._value is None:
            return default()

        return f(self._value)

    def filter(self, predicate: Fn[ValueT, bool]) -> Some[ValueT]:
        """Returns `Some[None]` if the contained value is `None`,

        otherwise calls the predicate and returns `Some[ValueT]` if the predicate returns `True`.

        Example
        -------
        ```py
        value = Some([1, 2, 3])

        print(value.filter(lambda x: len(x) >= 2))
        # Some([1, 2, 3])

        value: Some[int] = Some(None)
        print(value.filter(lambda x: x > 3))
        # None
        ```
        """
        if (value := self._value) is not None:
            if predicate(value):
                return Some(value)

        # We could return `None` here, but we want to return `Some[None]` instead.
        return Some(None)

    def take(self) -> None:
        """Take the value from the `Some` object setting it to `None`.

        Example
        -------
        ```py
        value = Some("Hi")
        value = value.take()
        print(value)
        # None
        ```
        """
        v = self._value = None
        return v

    def replace(self, value: ValueT) -> Some[ValueT]:
        """Replace the contained value with another value.

        Example
        -------
        ```py
        value = Some("Hi")
        value = value.replace("Hello")

        print(value)
        # Some("Hello")
        ```
        """
        v = self._value = value
        return Some(v)

    # Boolean checkings.
    def expect(self, message: str, /) -> ValueT:
        """Returns `ValueT` if the contained value is not `None` otherwise raises a `RuntimeError`.

        Example
        -------
        ```py
        value = Some("Hello")

        print(value.expect("Value is None"))
        # "Hello"

        value: Some[str] = Some(None)
        print(value.expect("Value is None"))
        # RuntimeError("Value is None")
        ```
        """
        if self._value is None:
            raise RuntimeError(message)

        return self._value

    def and_ok(self, optb: Some[T | None]) -> Some[T | None]:
        """Returns `Some[None]` if the contained value is `None`,

        Otherwise return optb as `Some[T | None]` if optb is `Some[T]`.

        Example
        -------
        ```py
        value = Some(5)

        print(value.and_ok(Some(10)))
        # Some(10)

        value: Some[int] = Some(10)
        print(value.and_ok(Some(None)))  # optb is `None`.
        # Some(None)
        ```
        """
        if self._value is None:
            return Some(None)

        return optb

    def and_then(self, f: Fn[ValueT, Some[T | None]]) -> Some[T | None]:
        """Returns `Some[None]` if the contained value is `None`,

        otherwise call `f` on `ValueT` and return `Some[T]` if it's value not `None`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.and_then(lambda x: Some(x * 2)))
        # Some(10)

        value: Some[int] = Some(10)
        print(value.and_then(lambda x: Some(None)))
        # Some(None)
        ```
        """
        if self._value is None:
            return Some(None)

        return f(self._value)

    def as_ref(self) -> Some[_ref.Ref[ValueT]]:
        """Returns immutable `Some[Ref[ValueT]]` if the contained value is not `None`,

        Otherwise returns `Some[None]`.

        Example
        -------
        ```py
        value = Some(5).as_ref().unwrap()
        value.object = 0 # FrozenError!

        owned = value.object
        print(owned) # 5

        # Create a copy of object.
        clone = value.copy()
        clone = 0  # Thats fine.
        print(clone == owned) # False, 0 != 5

        # None object.
        value: Some[int] = Some(None)
        print(value.as_ref())
        # Some(Ref(None))
        ```

        Raises
        ------
        `dataclasses.FrozenInstanceError`
            When attempting to modify the contained value. Use `sain.Ref.copy()` method to create a copy.

            Or just use `Some.as_ref_mut()` if you're dealing with mutable objects.
        """
        if self._value is not None:
            return Some(_ref.Ref(self._value))

        return Some(None)

    def as_mut(self) -> Some[_ref.RefMut[ValueT]]:
        """Returns mutable `Some[RefMut[ValueT]]` if the contained value is not `None`,

        Otherwise returns `Some[None]`.

        Example
        -------
        ```py
        value = Some(5).as_ref_mut().unwrap()
        value.object = 0
        print(value.object) # 0

        # None object.
        value: Some[int] = Some(None)
        print(value.as_ref_mut())
        # Some(RefMut(None))
        ```
        """
        if self._value is not None:
            return Some(_ref.RefMut(self._value))

        return Some(None)

    def is_some(self) -> bool:
        """Returns `True` if the contained value is not `None`, otherwise returns `False`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.is_some())
        # True

        value: Some[int] = Some(None)
        print(value.is_some())
        # False
        ```
        """
        return self._value is not None

    def is_some_and(self, predicate: Fn[ValueT, bool]) -> bool:
        """Returns `True` if the contained value is not `None` and
        the predicate returns `True`, otherwise returns `False`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.is_some_and(lambda x: x > 3))
        # True

        value: Some[int] = Some(None)
        print(value.is_some_and(lambda x: x > 3))
        # False
        ```
        """
        return self._value is not None and predicate(self._value)

    def is_none(self) -> bool:
        """Returns `True` if the contained value is `None`, otherwise returns `False`.

        Example
        -------
        ```py
        value = Some(5)
        print(value.is_none())
        # False

        value: Some[int] = Some(None)
        print(value.is_none())
        # True
        ```
        """
        return not self.is_some()

    def __str__(self) -> str:
        return f"Some({self._value!r})"

    def __repr__(self) -> str:
        return f"Some({self._value!r})"

    def __bool__(self) -> bool:
        return self.is_some()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Some):
            return NotImplemented

        return self._value == other.read

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self._value)


# This is declared as a typealias in .pyi
Option = Some[ValueT]
"""A type hint for a value that can be `Some[T]`.

Note
----
This must be used as a type hint for a value that can be `Some[T]`

and not for instance creation for `Some`.

Example
-------
```py
import sain

foo: sain.Option[str] = sain.Some(None)
```
"""



# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import typing as _typing

import typing_extensions as _typing_extensions

from . import default as _default
from . import ref as _ref

ValueT = _typing.TypeVar("ValueT")
# OkT = typing.TypeVar("OkT")
# ErrT = typing.TypeVar("ErrT")
T = _typing.TypeVar("T")
T_co = _typing.TypeVar("T_co", covariant=True)
Fn = _typing.Callable[[ValueT], T]
FnOnce = _typing.Callable[[], T]

Option: _typing_extensions.TypeAlias = "Some[ValueT]"

class Some(_typing.Generic[ValueT], _default.Default[None]):
    def __init__(self, value: ValueT | None) -> None: ...
    @staticmethod
    def default() -> None: ...
    @property
    def read(self) -> ValueT | None: ...
    def unwrap(self, /) -> ValueT | _typing.NoReturn: ...
    def unwrap_or(self, default: ValueT) -> ValueT: ...
    def unwrap_or_else(self, f: FnOnce[ValueT]) -> ValueT: ...
    def map(self, f: Fn[ValueT, T]) -> Some[T | None]: ...
    def map_or(self, default: T, f: Fn[ValueT, T]) -> T: ...
    def map_or_else(self, default: FnOnce[T], f: Fn[ValueT, T]) -> T: ...
    def filter(self, predicate: Fn[ValueT, bool]) -> Some[ValueT]: ...
    def take(self) -> None: ...
    def replace(self, value: ValueT) -> Some[ValueT]: ...
    def expect(self, message: str) -> ValueT: ...
    def and_ok(self, optb: Some[T | None]) -> Some[T | None]: ...
    def and_then(self, f: Fn[ValueT, Some[T | None]]) -> Some[T | None]: ...
    def as_ref(self) -> Some[_ref.Ref[ValueT]]: ...
    def as_mut(self) -> Some[_ref.RefMut[ValueT]]: ...
    def is_some(self) -> bool: ...
    def is_some_and(self, predicate: Fn[ValueT, bool]) -> bool: ...
    def is_none(self) -> bool: ...
    def __bool__(self) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...



import typing as _typing

import typing_extensions as _typing_extensions

Signature = _typing.TypeVar("Signature", bound=_typing.Callable[..., object])
TARGET_OS: _typing_extensions.TypeAlias = _typing.Literal["linux", "win32", "darwin", "unix", "windows"]
TARGET_ARCH: _typing_extensions.TypeAlias = _typing.Literal["x86", "x64", "arm", "arm64"]
PY_IMPL: _typing_extensions.TypeAlias = _typing.Literal["CPython", "PyPy", "IronPython", "Jython"]

def cfg_attr(
    *,
    requires_modules: str | _typing.Sequence[str] | None = ...,
    target_os: TARGET_OS | None = ...,
    python_version: tuple[int, int, int] | None = ...,
    target_arch: TARGET_ARCH | None = ...,
    impl: PY_IMPL | None = ...,
) -> _typing.Callable[[Signature], Signature]: ...
def cfg(
    *,
    target_os: TARGET_OS | None = ...,
    requires_modules: str | _typing.Sequence[str] | None = ...,
    python_version: tuple[int, int, int] | None = ...,
    target_arch: TARGET_ARCH | None = ...,
    impl: PY_IMPL | None = ...,
) -> bool: ...

class _AttrCheck(_typing.Generic[Signature]):
    def __init__(
        self,
        callback: Signature,
        target_os: TARGET_OS | None = ...,
        requires_modules: str | _typing.Sequence[str] | None = ...,
        python_version: tuple[int, int, int] | None = ...,
        target_arch: TARGET_ARCH | None = ...,
        impl: PY_IMPL | None = ...,
        *,
        no_raise: bool = ...,
    ) -> None: ...
    def __call__(self, *args: _typing.Any, **kwds: _typing.Any) -> Signature: ...
    def cfg_check(self) -> bool: ...



# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Module contains a Standard functional iterator implementation."""

from __future__ import annotations

__all__: typing.Tuple[str, str, str] = ("Iter", "into_iter", "Item")

import itertools
import typing

from . import default as _default

Item = typing.TypeVar("Item")
"""A type hint for the item type of the iterator."""

if typing.TYPE_CHECKING:
    import _typeshed as typeshed

    OtherItem = typing.TypeVar("OtherItem")
    _B = typing.TypeVar("_B", bound=typing.Callable[..., typing.Any])


class Iter(typing.Iterator[Item], typing.Generic[Item], _default.Default[typing.Iterable[Item]]):
    """Lazy, In-Memory iterator for sequence types with some functional methods.

    Example
    -------
    ```py
    iterator = Iter([1, 2, 3])
    # Limit the results to 2.
    for item in iterator.take(2):
        print(item)
    # 1
    # 2

    # Filter the results.
    for item in iterator.filter(lambda item: item > 1):
        print(item)
        print(iterator.count())
    # 2
    # 3
    # 3

    # Indexing is supported.
    print(iterator[0])
    # 1
    ```

    Parameters
    ----------
    items: `typing.Iterable[Item]`
        The items to iterate over. This must be an iterable.
    """

    __slots__ = ("_items",)

    def __init__(self, items: typing.Iterable[Item]) -> None:
        self._items = iter(items)

    @staticmethod
    def default() -> typing.Iterable[Item]:
        """Returns the default value for the iterator. An empty list is returned."""
        return []

    @typing.overload
    def collect(self, /) -> typing.List[Item]:
        ...

    @typing.overload
    def collect(self, casting: _B, /) -> typing.List[_B]:
        ...

    def collect(self, casting: typing.Optional[_B] = None, /) -> typing.Union[typing.List[Item], list[_B]]:
        """Collects all items in the iterator into a list.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> iterator.collect()
        [1, 2, 3]
        >>> iterator.collect(str)
        ['1', '2', '3']

        Parameters
        ----------
        casting: `T | None`
            An optional type to cast the items into.
            If not provided the items will be returned as a normal list.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        if casting is not None:
            return typing.cast("list[_B]", list(map(casting, self._items)))

        return list(self._items)

    def next(self) -> Item:
        """Returns the next item in the iterator.

        Example
        -------
        >>> iterator = Iter[str](["1", "2", "3"])
        item = iterator.next()
        assert item == "1"
        item = iterator.next()
        assert item == "2"

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        try:
            return self.__next__()
        except StopIteration:
            self._ok()

    def map(self, predicate: typing.Callable[[Item], OtherItem]) -> Iter[OtherItem]:
        """Maps each item in the iterator to its predicated value.

        Example
        -------
        >>> iterator = Iter[str](["1", "2", "3"]).map(lambda value: int(value))
        <Iter([1, 2, 3])>
        >>> for item in iterator:
                assert isinstance(item, int)

        Parameters
        ----------
        predicate: `typing.Callable[[Item], Item]`
            The function to map each item in the iterator to its predicated value.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(map(predicate, self._items))

    def take(self, n: int) -> Iter[Item]:
        """Take the first number of items until the number of items are yielded or
        the end of the iterator is reached.

        Example
        -------
        >>> iterator = Iter([RAID, STRIKE, GAMBIT])
        >>> for mode in iterator.take(2):
                assert mode in [RAID, STRIKE]
        <Iter([RAID, STRIKE])>

        Parameters
        ----------
        n: `int`
            The number of items to take.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(itertools.islice(self._items, n))

    def take_while(self, predicate: typing.Callable[[Item], bool]) -> Iter[Item]:
        """Yields items from the iterator while predicate returns `True`.

        Example
        -------
        ```py
        iterator = Iter(['VIP', 'Regular', 'Guard'])
        for membership in iterator.take_while(lambda m: m is 'VIP'):
            print(membership)

        # VIP
        ```

        Parameters
        ----------
        predicate: `typing.Callable[[Item], bool]`
            The function to predicate each item in the iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(itertools.takewhile(predicate, self._items))

    def drop_while(self, predicate: typing.Callable[[Item], bool]) -> Iter[Item]:
        """Yields items from the iterator while predicate returns `False`.

        Example
        -------
        ```py
        iterator = Iter(['VIP', 'Regular', 'Guard'])
        for membership in iterator.drop_while(lambda m: m is not 'VIP'):
            print(membership)

        # Regular
        # Guard
        ```

        Parameters
        ----------
        predicate: `typing.Callable[[Item], bool]`
            The function to predicate each item in the iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(itertools.dropwhile(predicate, self._items))

    def filter(self, predicate: typing.Callable[[Item], bool]) -> Iter[Item]:
        """Filters the iterator to only yield items that match the predicate.

        Example
        -------
        ```py
        places = Iter(['London', 'Paris', 'Los Angeles'])
        for place in places.filter(lambda place: place.startswith('L')):
            print(place)

        # London
        # Los Angeles
        ```
        """
        return Iter(filter(predicate, self._items))

    def skip(self, n: int) -> Iter[Item]:
        """Skips the first number of items in the iterator.

        Example
        -------
        ```py
        iterator = Iter([MembershipType.STEAM, MembershipType.XBOX, MembershipType.STADIA])
        for platform in iterator.skip(1):
                print(platform)
        # Skip the first item in the iterator.
        # <Iter([MembershipType.XBOX, MembershipType.STADIA])>
        """
        return Iter(itertools.islice(self._items, n, None))

    def discard(self, predicate: typing.Callable[[Item], bool]) -> Iter[Item]:
        """Discards all elements in the iterator for which the predicate function returns true.

        Example
        -------
        >>> iterator = Iter([MembershipType.STEAM, MembershipType.XBOX, MembershipType.STADIA])
        >>> for _ in iterator.discard(lambda platform: platform is not MembershipType.STEAM):
                # Drops all memberships that are not steam.
                print(iterator)
        <Iter([MembershipType.XBOX, MembershipType.STADIA])>

        Parameters
        ----------
        predicate: `typing.Callable[[Item], bool]`
            The function to test each item in the iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(filter(lambda x: not predicate(x), self._items))

    def zip(self, other: Iter[OtherItem]) -> Iter[typing.Tuple[Item, OtherItem]]:
        """Zips the iterator with another iterable.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> other = Iter([4, 5, 6])
        >>> for item, other_item in iterator.zip(other):
                assert item == other_item
        <Iter([(1, 4), (2, 5), (3, 6)])>

        Parameters
        ----------
        other: `Iter[OtherItem]`
            The iterable to zip with.

        Returns
        -------
        `Iter[tuple[Item, OtherItem]]`
            The zipped iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(zip(self._items, other))

    def all(self, predicate: typing.Callable[[Item], bool]) -> bool:
        """Return `True` if all items in the iterator match the predicate.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> while iterator.all(lambda item: isinstance(item, int)):
                print("Still all integers")
                continue
            # Still all integers

        Parameters
        ----------
        predicate: `typing.Callable[[Item], bool]`
            The function to test each item in the iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return all(predicate(item) for item in self)

    def any(self, predicate: typing.Callable[[Item], bool]) -> bool:
        """`True` if any items in the iterator match the predicate.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> if iterator.any(lambda item: isinstance(item, int)):
                print("At least one item is an int.")
        # At least one item is an int.

        Parameters
        ----------
        predicate: `typing.Callable[[Item], bool]`
            The function to test each item in the iterator.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return any(predicate(item) for item in self)

    def sort(
        self,
        *,
        key: typing.Callable[[Item], typeshed.SupportsRichComparison],
        reverse: bool = False,
    ) -> Iter[Item]:
        """Sorts the iterator.

        Example
        -------
        >>> iterator = Iter([3, 1, 6, 7])
        >>> for item in iterator.sort(key=lambda item: item < 3):
                print(item)
        # 1
        # 3
        # 6
        # 7

        Parameters
        ----------
        key: `typing.Callable[[Item], Any]`
            The function to sort by.
        reverse: `bool`
            Whether to reverse the sort.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(sorted(self._items, key=key, reverse=reverse))

    def first(self) -> Item:
        """Returns the first item in the iterator.

        Example
        -------
        >>> iterator = Iter([3, 1, 6, 7])
        >>> iterator.first()
        3

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return self.take(1).next()

    def reversed(self) -> Iter[Item]:
        """Returns a new iterator that yields the items in the iterator in reverse order.

        Example
        -------
        >>> iterator = Iter([3, 1, 6, 7])
        >>> for item in iterator.reversed():
                print(item)
        # 7
        # 6
        # 1
        # 3

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(reversed(self.collect()))

    def count(self) -> int:
        count = 0
        for _ in self:
            count += 1

        return count

    def union(self, other: Iter[Item]) -> Iter[Item]:
        """Returns a new iterator that yields all items from both iterators.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> other = Iter([4, 5, 6])
        >>> for item in iterator.union(other):
                print(item)
        # 1
        # 2
        # 3
        # 4
        # 5
        # 6

        Parameters
        ----------
        other: `Iter[Item]`
            The iterable to union with.

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(itertools.chain(self._items, other))

    def for_each(self, func: typing.Callable[[Item], typing.Any]) -> None:
        """Calls `func` on each item in the iterator.

        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> iterator.for_each(lambda item: print(item))
        # 1
        # 2
        # 3

        Parameters
        ----------
        func: `typing.Callable[[Item], typing.Any]`
            The function to call on each item in the iterator.
        """
        for item in self:
            func(item)

    def enumerate(self, *, start: int = 0) -> Iter[typing.Tuple[int, Item]]:
        """Returns a new iterator that yields tuples of the index and item.
        Example
        -------
        >>> iterator = Iter([1, 2, 3])
        >>> for index, item in iterator.enumerate():
                print(index, item)
        # 0, 1
        # 1, 2
        # 2, 3

        Raises
        ------
        `StopIteration`
            If no elements are left in the iterator.
        """
        return Iter(enumerate(self._items, start=start))

    def _ok(self) -> typing.NoReturn:
        raise StopIteration("No more items in the iterator.") from None

    def __getitem__(self, index: int) -> Item:
        try:
            return self.skip(index).first()
        except IndexError:
            self._ok()

    # This is a never.
    def __setitem__(self) -> typing.NoReturn:
        raise TypeError(f"{type(self).__name__} doesn't support item assignment.") from None

    def __contains__(self, item: Item) -> bool:
        return item in self._items

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({", ".join([str(item) for item in self._items])})>'

    def __len__(self) -> int:
        return self.count()

    def __iter__(self) -> Iter[Item]:
        return self

    def __next__(self) -> Item:
        try:
            item = next(self._items)
        except StopIteration:
            self._ok()

        return item


def into_iter(
    iterable: typing.Iterable[Item],
) -> Iter[Item]:
    """Convert an iterable into `Iter`.
    Example
    -------
    ```py
    sequence = [1,2,3]
    for item in sain.into_iter(sequence).reversed():
            print(item)
    # 3
    # 2
    # 1
    ```

    Parameters
    ----------
    iterable: `typing.Iterable[Item]`
        The iterable to convert.

    Raises
    ------
    `StopIteration`
        If no elements are left in the iterator.
    """
    return Iter(iterable)



# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Runtime attr confuguration."""

from __future__ import annotations

__all__: typing.Tuple[str, ...] = ("cfg_attr", "cfg")

import functools
import inspect
import platform
import sys
import typing

import pkg_resources

SigT = typing.Callable[..., object]
Signature = typing.TypeVar("Signature", bound=SigT)
"""A type var hint for the decorated object signature."""

TARGET_OS = typing.Literal["linux", "win32", "darwin", "unix", "windows"]
TARGET_ARCH = typing.Literal["x86", "x64", "arm", "arm64"]
PY_IMPL = typing.Literal["CPython", "PyPy", "IronPython", "Jython"]


def _machine() -> str:
    return platform.machine()


def _is_arm() -> bool:
    return _machine().startswith("arm")


def _is_arm_64() -> bool:
    return _is_arm() and _is_x64()


def _is_x64() -> bool:
    return _machine().endswith("64")


def cfg_attr(
    *,
    requires_modules: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
    target_os: typing.Optional[TARGET_OS] = None,
    python_version: typing.Optional[typing.Tuple[int, int, int]] = None,
    target_arch: typing.Optional[TARGET_ARCH] = None,
    impl: typing.Optional[PY_IMPL] = None,
) -> typing.Callable[[Signature], Signature]:
    """Conditional runtime object configuration based on passed arguments.

    If the decorated object gets called and one of the attributes returns `False`,
    `RuntimeError` will be raised and the object will not run.

    Example
    -------
    ```py
    import sain

    @sain.cfg_attr(target_os = "windows")
    def windows_only():
        # Do stuff with Windows's API.
        ...

    # Mut be PyPy Python implementation or `RuntimeError` will be raised
    # when creating the instance.
    @sain.cfg_attr(impl="PyPy")
    class Zoo:
        @sain.cfg_attr(target_os = "linux")
        def bark(self) -> None:
            windows_only()  # RuntimeError("Windows OS only!)

    # An instance will not be created if raised.
    zoo = Zoo()
    # RuntimError("class Zoo requires PyPy implementation")
    zoo.bark()
    # Whats zoo??
    ```

    Parameters
    ----------
    requires_modules : `str | Sequence[str] | None`
        A string or sequence of the required modules for the object to be ran.
    target_os : `str | None`
        The targeted operating system thats required for the object to be ran.
    python_version : `tuple[int, int, int] | None`
        The targeted Python version thats required for the object to be ran.

        Format must be `(3, 9, 5)`.
    target_arch : `str | None`
        The CPU targeted architecture thats required for the object to be ran.
    impl : `str | None`
        The Python implementation thats required for the object to be ran.

    Raises
    ------
    `RuntimeError`
        This fails if any of the attributes returns `False`. `required_modules` is not included.
    `ModuleNotFoundError`
        If the module check fails. i.e., if `required_modules` was provided and it returns `False`.
    `ValueError`
        If the passed Python implementation is unknown.
    """

    def decorator(callback: Signature) -> Signature:
        @functools.wraps(callback)
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> Signature:
            checker = _AttrCheck[Signature](
                callback,
                requires_modules=requires_modules,
                target_os=target_os,
                python_version=python_version,
                target_arch=target_arch,
                impl=impl,
            )
            return checker(*args, **kwargs)

        return typing.cast(Signature, wrapper)

    return decorator


def cfg(
    target_os: typing.Optional[TARGET_OS] = None,
    requires_modules: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
    python_version: typing.Optional[typing.Tuple[int, int, int]] = None,
    target_arch: typing.Optional[TARGET_ARCH] = None,
    impl: typing.Optional[PY_IMPL] = None,
) -> bool:
    """A function that will run the code only if all predicate attributes returns `True`.

    The difference between this function and `cfg_attr` is that this function will not raise an exception.
    Instead it will return `False` if any of the attributes fails.

    Example
    -------
    ```py
    import sain

    if sain.cfg(target_os = "windows"):
        print("Windows")
    elif sain.cfg(target_os = "linux"):
        print("Linux")
    else:
        print("Something else")
    ```

    Parameters
    ----------
    requires_modules : `str | Sequence[str] | None`
        A string or sequence of the required module names for the object to be ran.
    target_os : `str | None`
        The targeted operating system thats required for the object to be ran.
    python_version : `tuple[int, int, int] | None`
        The targeted Python version thats required for the object to be ran.

        Format must be `(3, 9, 5)`.
    target_arch : `str | None`
        The CPU targeted architecture thats required for the object to be ran.
    impl : `str | None`
        The Python implementation thats required for the object to be ran.

    Returns
    -------
    `bool`
        The condition that was checked.
    """
    checker = _AttrCheck(
        lambda: None,
        no_raise=True,
        requires_modules=requires_modules,
        target_os=target_os,
        python_version=python_version,
        target_arch=target_arch,
        impl=impl,
    )
    return checker.internal_check()


class _AttrCheck(typing.Generic[Signature]):
    __slots__ = (
        "_requires_modules",
        "_target_os",
        "_callback",
        "_py_version",
        "_no_raise",
        "_target_arch",
        "_py_impl",
    )

    def __init__(
        self,
        callback: Signature,
        requires_modules: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        target_os: typing.Optional[TARGET_OS] = None,
        python_version: typing.Optional[typing.Tuple[int, int, int]] = None,
        target_arch: typing.Optional[TARGET_ARCH] = None,
        impl: typing.Optional[PY_IMPL] = None,
        *,
        no_raise: bool = False,
    ) -> None:
        self._callback = callback
        self._requires_modules = requires_modules
        self._target_os = target_os
        self._py_version = python_version
        self._target_arch = target_arch
        self._no_raise = no_raise
        self._py_impl = impl

    def __call__(self, *args: typing.Any, **kwds: typing.Any) -> Signature:
        self._check_once()
        return typing.cast(Signature, self._callback(*args, **kwds))

    def internal_check(self) -> bool:
        return self._check_once()

    def _check_once(self) -> bool:
        results: list[bool] = []
        if self._target_os is not None:
            results.append(self._check_platform())

        if self._py_version is not None:
            results.append(self._check_py_version())

        if self._requires_modules is not None:
            results.append(self._check_modules())

        if self._target_arch is not None:
            results.append(self._check_target_arch())

        if self._py_impl is not None:
            results.append(self._check_py_impl())

        # No checks are passed to cfg(), We return False.
        if not results:
            return False

        return all(results)

    def _check_modules(self) -> bool:
        required_modules: set[str] = set()

        assert self._requires_modules
        if isinstance(modules := self._requires_modules, str):
            modules = (modules,)

        for module in modules:
            try:
                pkg_resources.get_distribution(module)
                required_modules.add(module)
            except pkg_resources.DistributionNotFound:
                if self._no_raise:
                    return False
                else:
                    needed = (mod for mod in modules if mod not in required_modules)
                    return self._raise_or_else(
                        ModuleNotFoundError(self._output_str(f"requires modules {', '.join(needed)} to be installed"))
                    )
        return True

    def _check_platform(self) -> bool:
        is_unix = sys.platform in {"linux", "darwin"}

        # If the target os is unix, then we assume that it's either linux or darwin.
        if self._target_os == "unix" and is_unix:
            return True

        # Alias to win32
        if self._target_os == "windows" and sys.platform == "win32":
            return True

        if sys.platform == self._target_os:
            return True

        return self._raise_or_else(RuntimeError(self._output_str(f"requires {self._target_os} OS")))

    def _check_py_version(self) -> bool:
        if self._py_version and self._py_version <= sys.version_info:
            return True

        return self._raise_or_else(
            RuntimeError(
                self._output_str(f"requires Python >={self._py_version}. But found {platform.python_version()}")
            )
        )

    def _check_target_arch(self) -> bool:
        if self._target_arch:
            if self._target_arch == "arm":
                return _is_arm()
            elif self._target_arch == "arm64":
                return _is_arm_64()
            elif self._target_arch == "x86":
                return not _is_x64()
            elif self._target_arch == "x64":
                return _is_x64()
            else:
                raise ValueError(f"Unknown target arch: {self._target_arch}")

        return False

    def _check_py_impl(self) -> bool:
        if platform.python_implementation() == self._py_impl:
            return True

        return self._raise_or_else(RuntimeError(self._output_str(f"requires Python implementation {self._py_impl}")))

    @property
    def _obj_type(self) -> str:
        if inspect.isfunction(self._callback):
            return "function"
        elif inspect.isclass(self._callback):
            return "class"

        return "object"

    def _output_str(self, message: str, /) -> str:
        fn_name = "" if self._callback.__name__ == "<lambda>" else self._callback.__name__
        return f"{self._obj_type} {fn_name} {message}."

    def _raise_or_else(self, exception: BaseException, /) -> bool:
        if self._no_raise:
            return False
        else:
            raise exception from None



from . import default as default
from . import iter as iter
from . import option as option
from . import ref as ref
from .cfg import cfg as cfg
from .cfg import cfg_attr as cfg_attr
from .default import Default as Default
from .drop import Drop as Drop
from .drop import drop as drop
from .iter import Iter as Iter
from .iter import into_iter as into_iter
from .option import Option as Option
from .option import Some as Some
from .ref import Ref as Ref
from .ref import RefMut as RefMut

__all__: tuple[str, ...]
__url__: str
__author__: str
__version__: str
__docs__: str
__license__: str



# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Protocol to provide more functionality when objects get Dropped(garbage collected)."""
from __future__ import annotations

__all__: tuple[str, str] = ("Drop", "drop")

import typing


@typing.runtime_checkable
class Drop(typing.Protocol):
    """Protocol that types can implement to provide more functionality when
    they get garbage collected.

    Example
    -------
    ```py
    import sain
    import requests
    import dataclasses

    @dataclasses.dataclass
    class Connector(sain.Drop):
        session = requests.Session()

        # Called internally.
        def drop(self) -> None:
            self.session.close()

        def on_drop(self) -> None:
            print("Session is closed...")

        def get(self, url: str) -> requests.Response:
            return self.session.get(url)

    cxn = Connector()
    response = cxn.get("https://example.com")
    print(response.status_code)
    # 200
    # Session is closed...
    ```
    """

    # Not slotted by default.
    __dropped: bool = False

    def drop(self) -> None:
        """Manually drop an object."""
        raise NotImplementedError("Method `drop` must be implemented.")

    def on_drop(self) -> None:
        """Register a callback to be called when the object is dropped."""
        if not self.__dropped:
            raise RuntimeError("Object not dropped yet.")

    def __del__(self) -> None:
        self.drop()
        self.__dropped = True
        self.on_drop()


def drop(obj: Drop) -> None:
    """Drop objects that implements `Drop`."""
    obj.drop()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sain']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'sain',
    'version': '0.0.3',
    'description': 'Standard Rust core types implementations for Python.',
    'long_description': '# sain\nA pure Python package that implements standard Rust core types for Python.\n\n\n## Install\nYou\'ll need Python 3.8 or higher.\n\nPyPI\n```rs\n$ pip install sain\n```\n\n## Examples\nMore stuff in [examples](https://github.com/nxtlo/sain/tree/master/examples)\n\n### `cfg`, `cfg_attr`\nConditionally include code.\n\n```py\nfrom sain import cfg, cfg_attr\n\n@cfg_attr(target_os="unix")\n# Calling this on a non-unix system will raise a RuntimeError\n# and the function will not run.\ndef run_when_unix() -> None:\n    import uvloop\n    uvloop.install()\n\nif cfg(target_arch="arm64"):\n    run_when_unix()\n\n# If this returns True, Function will be in scope.\nif cfg(requires_modules="hikari-tanjun"):\n    def create_client() -> tanjun.Client:\n        return tanjun.Client(...)\n\n# Assuming tanjun is not installed.\n# Calling the function will raise `NameError`\n# since its not in scope.\ncreate_client()\n```\n\n### `Option<T>` and `Some<T>`\nImplements the standard `Option` and `Some` types. An object that may be `None` or `T`.\n\n```py\nimport sain\nimport os\n\n# Stright up replace typing.Optional[str]\ndef get_token(key: str) -> sain.Option[str]:\n    # What os.getenv returns may be str or None.\n    return sain.Some(os.getenv(key))\n\n# Raises RuntimeError("No token found.") if os.getenv return None.\ntoken = get_token().expect("No token found.")\n\n# The classic way to handle this in Python would be.\nif token is None:\n    token = "..."\nelse:\n    ...\n\n# Replace this with `unwrap_or`.\n# Returning DEFAULT_TOKEN if it was None.\nenv_or_default = get_token().unwrap_or("DEFAULT_TOKEN")\n\n# Safe type hint is fine.\nas_none: sain.Option[str] = sain.Some(None)\nas_none.uwnrap_or(123)  # Error: Must be type `str`!\nassert as_none.is_none()\n```\n\n### Defaults\nA protocol that types can implement which have a default value.\n\n```py\nimport sain\nimport requests\n\nclass Session(sain.Default[requests.Session]):\n    # One staticmethod must be implemented and must return the same type.\n    @staticmethod\n    def default() -> requests.Session:\n        return requests.Session()\n\nsession = Session.default()\n```\n\n### Iter\nTurns normal iterables into `Iter` type.\n\n```py\nimport sain\n\nf = sain.Iter([1,2,3])\n# or f = sain.into_iter([1,2,3])\nassert 1 in f\n\nfor item in f.take_while(lambda i: i > 1):\n    print(item)\n```\n\n### Why\nThis is whats Python missing.\n\n### Notes\nSince Rust is a compiled language, Whatever predict in `cfg` and `cfg_attr` returns False will not compile.\n\nBut there\'s no such thing as this in Python, So `RuntimeError` will be raised and whatever was predicated will not run.\n',
    'author': 'nxtlo',
    'author_email': 'dhmony-99@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nxtlo/sain',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# BSD 3-Clause License
#
# Copyright (c) 2022-Present, nxtlo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Standard Rust core types implementations for Python.

Equavilant types
----------------
- `Option<T>` -> `sain.Option[T]` | `sain.Some[T]`
- `Result<T, E>` -> `sain.Result[T, E]`. Not implemented yet.
- `Default<T>` -> `sain.Default[T]`
- `AsRef<T>` -> `sain.Ref[T]`.
- `AsMut<T>` -> `sain.RefMut[T]`.
- `Iter<Item>` -> `sain.Iter[Item]`

Equavilant macros
-----------------
As decorators.

- `cfg!()` -> `sain.cfg`.
- `#[cfg_attr]` -> `sain.cfg_attr`.

Examples
--------
```py
import sain

# If a non windows machine runs this function, it will raise an error.
@sain.cfg_attr(target_os = "windows")
def windows_only() -> sain.Option[int]:
    return sain.Some(1)

@sain.cfg_attr(requires_modules="uvloop", target_os = "unix")
def run_uvloop() -> None:
    import uvloop
    uvloop.install()

@sain.cfg_attr(python_version = (3, 5, 0))
class Foo:

    @staticmethod
    @sain.cfg_attr(requires_modules = ("numpy", "pandas"))
    async def bar() -> None:
        ...
```

Notes
-----
Target OS must be one of the following:
* `linux`
* `win32` | `windows`
* `darwin`
* `unix`, which is assumed to be either linux or darwin.

Target architecture must be one of the following:
* `x86`
* `x64`
* `arm`
* `arm64`

Target Python implementation must be one of the following:
* `CPython`
* `PyPy`
* `IronPython`
* `Jython`
"""
from __future__ import annotations

__all__ = (
    # cfg.py
    "cfg",
    "cfg_attr",
    # default.py
    "Default",
    "default",
    # ref.py
    "Ref",
    "RefMut",
    "ref",
    # option.py
    "Some",
    "Option",
    "option",
    # iter.py
    "into_iter",
    "Iter",
    "iter",
    # drop.py
    "drop",
    "Drop",
)

# Module top level. Required for pdoc.
from . import default
from . import iter
from . import option
from . import ref
from .cfg import cfg
from .cfg import cfg_attr
from .default import Default
from .drop import Drop
from .drop import drop
from .iter import Iter
from .iter import into_iter
from .option import Option
from .option import Some
from .ref import Ref
from .ref import RefMut

__version__: str = "0.0.3"
__url__: str = "https://github.com/nxtlo/sain"
__author__: str = "nxtlo"
__about__: str = "A Rust like cfg attribs checking for Python."
__docs__: str = ""
__license__: str = "BSD 3-Clause License"
